# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2021)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import shelve as shlv
import sys as sstm
from pathlib import Path as path_t
from typing import Any, Union


def CacheDictionaryForMain(
    main: Union[str, path_t],
    /,
    *,
    folder_name: str = "_runtime",
    folder_postfix: str = "",
    cache_postfix: str = "",
) -> shlv.Shelf:
    """"""
    if isinstance(main, str):
        main = path_t(main)

    cache_folder = main.parent / (folder_name + folder_postfix)
    if not cache_folder.exists():
        cache_folder.mkdir()
    elif not cache_folder.is_dir():
        print(f"{cache_folder}: Not a suitable cache folder")
        sstm.exit(-1)

    return shlv.open(str(cache_folder / (main.stem + cache_postfix)))


def CloseCache(cache: Union[shlv.Shelf, Any]) -> None:
    """"""
    if isinstance(cache, shlv.Shelf):
        cache.close()
